# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2024 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the files 'R/topTags.R' of the Bioconductor edgeR package (version 3.38.4).

import logging
import numpy as np
import pandas as pd
from statsmodels.stats.multitest import multipletests

from .DGEExact import DGEExact


class TopTags:
    def __init__(self, table, adjust_method, comparison, test):
        self.table = table
        self.adjust_method = adjust_method
        self.comparison = comparison
        self.test = test

    def __repr__(self):
        return f"Coefficient: {self.comparison}\n{self.table}"


def topTags(self, n=10, adjust_method="fdr_bh", sort_by="PValue", p_value=1):
    """
    Extract the most differentially expressed genes (or sequence tags) from a test object, ranked either by *p*-value or by absolute log-fold-change.

    This function accepts a test statistic object created by any of the
    functions :func:`exactTest`, :func:`glmLRT`, :func:`glmTreat` or
    :func:`glmQLFTest` and extracts a readable dataframe of the most
    differentially expressed genes.  The dataframe collates the annotation and
    differential expression statistics for the top genes. The dataframe is
    wrapped in a :class:`TopTags` object that records the test statistic used
    and the multiple testing adjustment method.

    :func:`topTags` permits ranking by fold-change but the authors do not
    recommend fold-change ranking or fold-change cutoffs for routine RNA-Seq
    analysis. The *p*-value ranking is intended to be more biologically
    meaningful, especially if the *p*-values were computed using
    :func:`glmTreat`.

    Arguments
    ---------
    self : DGEExact or DGELRT
        object containing test statistics and *p*-values
    n : int
        maximum number of genes/tags to return
    adjust_method : str
        specify the method used to adjust *p*-values for multiple testing. See
        :func:`statsmodels.stats.multitest` for possible values
    sort_by : {"PValue, "logFC", "none"}
        specify the sort method

        - :code:`"PValue"` to sort by *p*-value
        - :code:`"logFC"` to sort by absolute log-fold-change
        - :code:`"none"` for no sorting

    p_value : float
        cutoff value for adjusted *p*-values. Only tags with adjusted
        *p*-values equal or lower than specified are returned.

    Returns
    -------
    TopTags
        an object with the following components:

        - :code:`table`, a dataframe containing differential expression results
          for the top genes in a sorted order. The number of rows is the
          smaller of :code:`n` and the number of genes with adjusted *p*-value
          less than or equal to :code:`p_value`. The dataframe includes all the
          annotation columns from :code:`self.genes` and all statistic columns
          from :code:`self.table` plus one of:

            - :code:`"FDR"`, false discovery rate (only when
              :code:`adjust_method` is :code:`"fdr_bh"`, :code:`"fdr_by"`))
            - :code:`"FWER"`, family-wise error rate (only when
              :code:`adjust_method` is :code:`"holm"`, :code:`"simes-hochberg"`,
              :code:`"hommel"` or :code:`"bonferroni"`)

        - :code:`adjust_method`, string specifying the method used to adjust
          *p*-values for multiple testing, same as input argument
        - :code:`comparison` the names of the two groups being compared (for
          :class:`DGEExact` objects) or the glm contrast being tested (for
          :class:`DGELRT` objects).
        - :code:`test`, string stating the name of the test
    """
    if self.table is None:
        raise ValueError("Need to run exactTest or glmLRT first")
    if isinstance(self, DGEExact):
        test = "exact"
    else:
        test = "glm"
    MultipleContrasts = test == "glm" and self.table.shape[1] > 4

    # Check n
    n = np.min([n, self.table.shape[0]])
    if n < 1:
        raise ValueError("No rows to output")

    # Check adjust_method
    FWER_methods = ["holm", "simes-hochberg", "hommel", "bonferroni"]
    FDR_methods = ["fdr_bh", "fdr_by"]
    if adjust_method not in FWER_methods and adjust_method not in FDR_methods:
        raise ValueError(f"invalid argument {adjust_method} for 'adjust_method'")

    # Check sort_by
    if sort_by == "p_value":
        sort_by = "PValue"
    if sort_by not in ["none", "PValue", "logFC"]:
        raise ValueError(f"invalid value {sort_by} for 'sort_by'")

    # Absolute log fold change
    if MultipleContrasts:
        if sort_by == "logFC":
            logging.warnings.warn(
                "Two or more logFC columns in DGELRT object. First logFC column used to rank by logFC"
            )
        alfc = np.abs(self.table["logFC"].iloc[:, 0])
    else:
        alfc = np.abs(self.table["logFC"])

    # Choose top genes
    if sort_by == "logFC":
        o = np.argsort(-alfc)
    elif sort_by == "PValue":
        o = np.argsort(
            np.array(
                [(a, b) for a, b in zip(self.table["PValue"], -alfc)],
                dtype=[("x", self.table["PValue"].dtype), ("y", alfc.dtype)],
            ),
            order=("x", "y"),
        )
    elif sort_by == "none":
        o = np.arange(self.table.shape[0])
    else:
        raise ValueError(f"invalid value {sort_by} for 'sort_by'")
    tab = self.table.iloc[o, :]

    # Add adjusted p-values if appropriate
    adj_p_val = multipletests(self.table["PValue"], method=adjust_method)[1]
    if adjust_method != "none":
        if adjust_method in FWER_methods:
            adjustment = "FWER"
        if adjust_method in FDR_methods:
            adjustment = "FDR"
        self.table.loc[o, adjustment] = adj_p_val[o]

    # Add gene annotation if appropriate
    if self.genes is not None:
        rn = self.genes.index
        for c in self.genes.columns:
            tab[c] = self.genes[c]
        tab.index = rn

    # Thin out fit p_value threshold
    if p_value < 1:
        sig = adj_p_val[o] <= p_value
        sig[np.isnan(sig)] = False
        tab = tab[sig, :]

    # Enough rows left?
    if tab.shape[0] < n:
        n = tab.shape[0]
    if n < 1:
        return pd.DataFrame()

    # Output object
    return TopTags(
        table=tab.iloc[:n, :],
        adjust_method=adjust_method,
        comparison=self.comparison,
        test=test,
    )



# distutils: language = c++
#-----------------------------------------------------------------------------
# Copyright (C) 2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------------

cdef extern from "matrix.h":
    cpdef void init_numpy_c_api()

# initialization
init_numpy_c_api()




import unittest
import numpy as np

from inmoose.limma import squeezeVar


class Test(unittest.TestCase):
    def test_squeezeVar(self):
        with self.assertRaisesRegex(ValueError, expected_regex="var is empty"):
            squeezeVar(np.array([]), df=3)

        res = squeezeVar(np.array([1]), df=3)
        self.assertEqual(res["df_prior"], 0)
        self.assertTrue(np.array_equal(res["var_prior"], [1]))
        self.assertTrue(np.array_equal(res["var_post"], [1]))

        res = squeezeVar(np.array([0.5, 1.3, 7.8]), df=3)
        self.assertAlmostEqual(res["df_prior"], 2.830777, places=6)
        self.assertAlmostEqual(res["var_prior"], 1.676927, places=6)
        self.assertTrue(
            np.allclose(res["var_post"], [1.071385, 1.482994, 4.827317], atol=1e-6)
        )

        res = squeezeVar(
            np.array([0.5, 0.5, 1.3, 1.3, 7.8, 7.8]),
            df=20,
            covariate=[1, 2, 3, 4, 5, 6],
        )
        self.assertAlmostEqual(res["df_prior"], 15.44626, places=5)
        self.assertTrue(
            np.allclose(
                res["var_prior"],
                [0.4322342, 0.6426360, 1.0331707, 1.9359150, 4.2277336, 9.9836662],
                atol=1e-6,
            )
        )
        self.assertTrue(
            np.allclose(
                res["var_post"],
                [0.4704700, 0.5621558, 1.1837250, 1.5771098, 6.2433295, 8.7515664],
                atol=1e-6,
            )
        )



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2024 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/residDF.R' of the Bioconductor edgeR package (version 3.38.4).

import numpy as np


def _comboGroups(truths):
    """
    Function that returns a list of lists of indices, where each vector refers to the rows with the same combination of True/False values in :code:`truths`.

    Arguments
    ---------
    truths : array_like
        Boolean matrix

    Returns
    -------
    list of lists of ints
        each element of this list is the list of row indices of :code:`truths`
        that contain the same combination of :code:`True`/:code:`False`
    """

    d = {}
    for i in range(truths.shape[0]):
        k = tuple(truths[i, :])
        if k not in d:
            d[k] = []
        d[k].append(i)

    return [v for v in d.values()]


def _residDF(zero, design):
    """
    Effective residual degrees of freedom after adjusting for exact zeros

    Arguments
    ---------
    zero : array_like
        matrix of boolean indicating the zero counts
    design : array_like
        design matrix

    Returns
    -------
    ndarray
        effective residual degrees of freedom, one element per row in :code:`zero`
    """
    nlib = zero.shape[1]
    ncoef = design.shape[1]
    nzero = zero.sum(axis=1)

    # Default is no zero
    DF = np.full(len(nzero), nlib - ncoef)

    # All zero case
    DF[nzero == nlib] = 0

    # Anything in between?
    somezero = (nzero > 0) & (nzero < nlib)
    if somezero.any():
        zero2 = zero[somezero, :]
        groupings = _comboGroups(zero2)

        # Identifying the true residual d.f. for each of these rows.
        DF2 = nlib - nzero[somezero]
        for i in groupings:
            zeroi = zero2[i[0], :]
            DF2[i] -= np.linalg.matrix_rank(design[~zeroi, :])
        DF2 = np.maximum(DF2, 0)
        DF[somezero] = DF2

    return DF



import unittest
import numpy as np
from scipy.stats import nbinom

from inmoose.edgepy import q2qnbinom


class Test(unittest.TestCase):
    def test_q2qnbinom(self):
        x = np.arange(20, dtype=float)
        input_mean = 10.0
        output_mean = 20.0
        dispersion = 0.1

        ref = np.array(
            [
                1.339746,
                3.720337,
                5.706665,
                7.602321,
                9.448944,
                11.2636,
                13.05531,
                14.82956,
                16.58996,
                18.33906,
                20.07871,
                21.81033,
                23.53501,
                25.25363,
                26.96691,
                28.67544,
                30.37969,
                32.0801,
                33.777,
                35.47071,
            ]
        )

        res = q2qnbinom(
            x, input_mean=input_mean, output_mean=output_mean, dispersion=dispersion
        )
        self.assertTrue(np.allclose(res, ref))



from .addPriorCount import addPriorCount
from .adjustedProfileLik import adjustedProfileLik
from .aveLogCPM import aveLogCPM
from .binomTest import binomTest
from .DGEGLM import DGEGLM
from .DGEList import DGEList
from .dispCoxReid import dispCoxReid
from .dispCoxReidInterpolateTagwise import dispCoxReidInterpolateTagwise
from .estimateGLMCommonDisp import estimateGLMCommonDisp
from .estimateGLMTagwiseDisp import estimateGLMTagwiseDisp
from .exactTest import exactTest
from .exactTestBetaApprox import exactTestBetaApprox
from .exactTestByDeviance import exactTestByDeviance
from .exactTestBySmallP import exactTestBySmallP
from .exactTestDoubleTail import exactTestDoubleTail
from .glmFit import glmFit, glmLRT
from .glmQLFit import glmQLFit, glmQLFTest, plotQLDisp
from .maximizeInterpolant import maximizeInterpolant
from .mglmLevenberg import mglmLevenberg
from .mglmOneGroup import mglmOneGroup
from .mglmOneWay import mglmOneWay, designAsFactor
from .movingAverageByCol import movingAverageByCol
from .nbinomDeviance import nbinomDeviance
from .predFC import predFC
from .q2qnbinom import q2qnbinom
from .splitIntoGroups import splitIntoGroups
from .stats import pnbinom, qnbinom
from .systematicSubset import systematicSubset
from .topTags import topTags
from .validDGEList import validDGEList
from .edgepy_cpp import *



# distutils: language = c++
#-----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2022-2024 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------------

# This file contains Cython ports of the original C++ code from various files of
# the Bioconductor edgeR package (version 3.38.4):
# - 'src/nbdev.cpp' and 'src/R_compute_nbdev.cpp' (function
#   `compute_unit_nb_deviance`)
# - 'R/q2qnbinom.R' (function _q2qnbinom)

import numpy as np
cimport cython
from libcpp.cmath cimport log
from scipy.linalg.cython_lapack cimport dormqr, dgeqrf, dtrtrs, dpotrf, dpotrs, dsytrf
from scipy.linalg.cython_blas cimport dgemv
from libc.math cimport sqrt
from scipy.special cimport cython_special as sp

cdef public object make_levenberg_result "make_levenberg_result"(
        ndarray[double, ndim=2] coefficients,
        ndarray[double, ndim=2] fitted_values,
        vector.vector[double] deviance,
        vector.vector[long] iter,
        vector.vector[char] failed):
    return (coefficients, fitted_values, deviance, iter, failed)

cdef public ndarray[double, ndim=1] vector2ndarray "vector2ndarray"(const vector.vector[double]& data):
    return np.asarray(data, order='F')

cdef char side = b'L'
cdef char trans_ormqr = b'T'
cdef char uplo = b'U'
cdef char trans_trtrs = b'N'
cdef char diag = b'N'
cdef int unity = 1

cdef public void build_QRdecomposition(QRdecomposition* self):
    # Setting up the workspace for dgeqrf
    cdef double tmpwork
    dgeqrf(&self.NR, &self.NC, self.Xcopy.data(), &self.NR, self.tau.data(), &tmpwork, &self.lwork_geqrf, &self.info)

    # Loading up the optimal WORK
    self.lwork_geqrf = <int>(tmpwork + 0.5)
    if self.lwork_geqrf < 1:
        self.lwork_geqrf = 1
    self.work_geqrf.resize(self.lwork_geqrf)

    # Repeating for dormqr
    dormqr(&side, &trans_ormqr, &self.NR, &unity, &self.NC, self.Xcopy.data(), &self.NR, self.tau.data(), self.effects.data(), &self.NR, &tmpwork, &self.lwork_ormqr, &self.info)

    self.lwork_ormqr = <int>(tmpwork + 0.5)
    if self.lwork_ormqr < 1:
        self.lwork_ormqr = 1
    self.work_ormqr.resize(self.lwork_ormqr)

cdef void decompose_QRdecomposition(QRdecomposition* self):
    dgeqrf(&self.NR, &self.NC, self.Xcopy.data(), &self.NR, self.tau.data(), self.work_geqrf.data(), &self.lwork_geqrf, &self.info)

cdef void solve_QRdecomposition(QRdecomposition* self):
    dormqr(&side, &trans_ormqr, &self.NR, &unity, &self.NC, self.Xcopy.data(), &self.NR, self.tau.data(), self.effects.data(), &self.NR, self.work_ormqr.data(), &self.lwork_ormqr, &self.info)
    if self.info:
        raise RuntimeError("Q**T multiplication failed")

    dtrtrs(&uplo, &trans_trtrs, &diag, &self.NC, &unity, self.Xcopy.data(), &self.NR, self.effects.data(), &self.NR, &self.info)
    if self.info:
        raise RuntimeError("failed to solve the triangular system")

cdef public void f77_dgemv "f77_dgemv"(const char *trans, const int *m, const int *n, const double *alpha, const double *a, const int *lda, const double *x, const int *incx, const double *beta, double *y, const int *incy) nogil:
    dgemv(<char*>trans, <int*>m, <int*>n, <double*>alpha, <double*>a, <int*>lda, <double*>x, <int*>incx, <double*>beta, y, <int*>incy)
cdef public void f77_dpotrf "f77_dpotrf"(const char *uplo, const int *n, double *a, const int *lda, int *info) nogil:
    dpotrf(<char*>uplo, <int*>n, a, <int*>lda, info)
cdef public void f77_dpotrs "f77_dpotrs"(const char *uplo, const int *n, const int *nrhs, const double *a, const int *lda, double *b, const int *ldb, int *info) nogil:
    dpotrs(<char*>uplo, <int*>n, <int*>nrhs, <double*>a, <int*>lda, b, <int*>ldb, info)
cdef public void f77_dsytrf "f77_dsytrf"(const char *uplo, const int *n, double *a, const int *lda, int *ipiv, double *work, const int *lwork, int *info) nogil:
    dsytrf(<char*>uplo, <int*>n, a, <int*>lda, ipiv, work, <int*>lwork, info)


cdef double mildly_low_value = 1e-8
@cython.ufunc
@cython.cdivision(True)
cdef double compute_unit_nb_deviance(double y, double mu, double phi):
    """
    Calculate the deviance of a negative binomial fit

    Note the protection for very large mu*phi (where we use a Gamma instead) or
    very small mu*phi (where we use a Poisson instead). This approximation
    protects against numerical instability introduced by subtracting a very
    large log value in (log mu) with another very large logarithm (log mu+1/phi).
    We need to consider the phi as the approximation is only good when the
    product is very big or very small.

    Arguments
    ---------
    y : array_like
        counts matrix
    mu : array_like
        expected means matrix (broadcastable to the shape of :code:`y`)
    phi : array_like
        dispersion matrix (broadcastable to the shape of :code:`y`)

    Returns
    -------
    ndarray
        matrix of deviances (same shape as :code:`y`)
    """
    # add a small value to protect against zero during division and log
    y = y + mildly_low_value
    mu = mu + mildly_low_value

    # Calculating the deviance using either the Poisson (small phi*mu), the
    # Gamma (large phi*mu) or NB (everything else). Some additional work is
    # put in to make the transitiosn between families smooth
    resid = y - mu
    product = mu * phi

    if phi < 1e-4:
        return 2 * (y * log(y/mu) - resid - 0.5*resid*resid*phi*(1+phi*(2/3*resid-y)))
    elif product > 1e6:
        return 2 * (resid/mu - log(y/mu)) * mu/(1+product)
    else:
        return 2 * (y * log(y/mu) + (y + 1/phi) * log((mu + 1/phi)/(y + 1/phi)))

@cython.ufunc
cdef double _q2qnbinom(double x, double input_mean, double output_mean, double dispersion):
    """
    Interpolated quantile to quantile mapping between negative-binomial distributions with the same dispersion but different means.

    This is the low-level Cythonized function called by :func:`q2qnbinom`.

    See also
    --------
    q2qnbinom

    Arguments
    ---------
    x : array_like
        matrix of counts
    input_mean : array_like
        matrix of population means for :code:`x`. If 1-D, then of the same
        length as :code:`x.shape[0]`
    output_mean : array_like
        matrix of population means for the output values. If 1-D, then of the
        same length as :code:`x.shape[0]`
    dispersion : array_like
        scalar, vector or matrix giving negative binomial dispersion values

    Returns
    -------
    ndarray
        matrix of same dimensions as :code:`x`, with :code:`output_mean` as the
        new nominal population mean
    """
    if x < 0:
        raise ValueError("x must be non-negative")
    if input_mean < 0:
        raise ValueError("input_mean must be non-negative")
    if output_mean < 0:
        raise ValueError("output_mean must be non-negative")
    if dispersion < 0:
        raise ValueError("dispersion must be non-negative")

    eps = 1e-14
    if input_mean < eps or output_mean < eps:
        input_mean += 0.25
        output_mean += 0.25
    ri = 1 + dispersion * input_mean
    vi = input_mean * ri
    ro = 1 + dispersion * output_mean
    vo = output_mean * ro

    q1 = output_mean + sqrt(vo / vi) * (x - input_mean)

    if x >= input_mean:
        p2 = sp.gammaincc(input_mean / ri, x/ri)
        q2 = sp.gammainccinv(output_mean / ro, p2) * ro
    else:
        p2 = sp.gammainc(input_mean / ri, x/ri)
        q2 = sp.gammaincinv(output_mean / ro, p2) * ro

    return (q1+q2)/2



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2024 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the files 'R/glmQLFTest.R' and 'src/R_check_poisson_bound.cpp' of the Bioconductor edgeR package (version 3.38.4).

import numpy as np
import patsy
import scipy

from .aveLogCPM import aveLogCPM
from .DGEGLM import DGEGLM
from .glmFit import glmFit, glmLRT
from .residDF import _residDF
from ..limma import squeezeVar


def glmQLFit_DGEList(
    self,
    design=None,
    dispersion=None,
    abundance_trend=True,
    robust=False,
    winsor_tail_p=(0.05, 0.1),
):
    """
    Fit a quasi-likelihood negative binomial generalized log-linear model to count data.

    See also
    --------
    glmQLFit

    Arguments
    ---------
    design : matrix, optional
        design matrix for the genewise linear models
    dispersion : float or array_like
        scalar, vector or matrix of negative binomial dispersions. If
        :code:`None`, it will be extracted from the :class:`~DGEList` object,
        with order of precedence: trended dispersions, common dispersion, a
        constant value of 0.05.
    abundance_trend : bool
        whether to allow an abundance-dependent trend when estimating the prior
        values for the quasi-likelihood multiplicative dispersion parameter.
    robust : bool
        whether to estimate the prior QL dispersion distribution robustly
    winsor_tail_p : pair of floats
        pair of floats giving the proportion to trim (Winsorize) from lower and
        upper tail of the distribution of genewise deviances when estimating
        the hyperparameters. Positive values produce robust empirical Bayes
        ignoring outlier small or large deviances. Only used when
        :code:`robust=True`.

    Returns
    -------
    DGEGLM
        object with the same components as produced by :func:`glmFit`, plus:

        - :code:`df_residual_zeros`, an array containing the number of
          effective residual degrees of freedom for each gene, taking into
          account any treatment groups with all zero counts.
        - :code:`df_prior`, a float (if :code:`robust=False`) or array (if
          :code:`robust=True`), giving the prior degrees of freedom for the QL
          dispersions.
        - :code:`var_prior`, a float (if :code:`robust=False`) or array (if
          :code:`robust=True`), giving the location of the prior distrbution
          for the QL dispersions.
        - :code:`var_post`, an array containing the posterior empirical Bayes
          QL dispersions.
    """

    if design is None:
        design = self.design
        if design is None:
            design = patsy.dmatrix("~C(self.samples['group'])")

    if dispersion is None:
        dispersion = self.trended_dispersion
        if dispersion is None:
            dispersion = self.common_dispersion
        if dispersion is None:
            raise ValueError("No dispersion values found in DGEList object")

    offset = self.getOffset()
    if self.AveLogCPM is None:
        self.AveLogCPM = self.aveLogCPM()

    fit = glmQLFit(
        y=self.counts,
        design=design,
        dispersion=dispersion,
        offset=offset,
        lib_size=None,
        abundance_trend=abundance_trend,
        AveLogCPM=self.AveLogCPM,
        robust=robust,
        winsor_tail_p=winsor_tail_p,
        weights=self.weights,
    )
    fit.samples = self.samples
    fit.genes = self.genes
    fit.AveLogCPM = self.AveLogCPM
    return fit


def glmQLFit(
    y,
    design=None,
    dispersion=None,
    offset=None,
    lib_size=None,
    weights=None,
    abundance_trend=True,
    AveLogCPM=None,
    robust=False,
    winsor_tail_p=(0.05, 0.1),
):
    """
    Fit a quasi-likelihood negative binomial generalized log-linear model to count data.

    Implement one of the quasi-likelihood (QL) methods of [Lund2012]_, with some
    enhancements and with slightly different glm, trend and FDR methods. See
    [Lun2016]_ or [Chen2016]_ for tutorials describing the use of
    :func:`glmQLFit` and :func:`glmQLFTest` as part of a complete pipeline.
    Another case study using :func:`glmQLFit` and :func:`glmQLFTest` is given in
    Section 4.7 of the edgeR User's Guide.

    :func:`glmQLFit` is similar to :func:`glmFit` except that it also estimates
    QL dispersion values. It calls the limma function :func:`.squeezeVar` to
    conduct empirical Bayes moderation of the genewise QL dispersions. If
    :code:`robust=True`, then the robust hyperparameter estimation features of
    :func:`.squeezeVar` are used [Phipson2016]_. If
    :code:`abundance_trend=True`, then a prior trend is estimated based on the
    average logCPMs.

    :func:`glmQLFit` gives special attention to handling of zero counts, and in
    particular to situations when fitted values of zero provide no useful
    residual degrees of freedom for estimating the QL dispersion [Lun2017]_.
    The usual residual degrees of freedom are returned as :code:`df_residual`
    while the adjusted residual degrees of freedom are returned as
    :code:`df_residual_zeros`.

    Note
    ----
    The negative binomial dispersions :code:`dispersion` supplied to
    :func:`glmQLFit` and :func:`glmQLFTest` must be based on a global model, that
    is, they must be either trended or common dispersions. It is not correct to
    supply genewise dispersions because :func:`glmQLFTest` estimates genewise
    variability using the QL dispersion.

    Arguments
    ---------
    y : matrix
        a matrix of counts
    design : matrix, optional
        design matrix for the genewise linear models
    dispersion : float or array_like
        scalar, vector or matrix of negative binomial dispersions. If
        :code:`None`, it will be extracted from the :class:`~DGEList` object,
        with order of precedence: trended dispersions, common dispersion, a
        constant value of 0.05.
    abundance_trend : bool
        whether to allow an abundance-dependent trend when estimating the prior
        values for the quasi-likelihood multiplicative dispersion parameter.
    robust : bool
        whether to estimate the prior QL dispersion distribution robustly
    winsor_tail_p : pair of floats
        pair of floats giving the proportion to trim (Winsorize) from lower and
        upper tail of the distribution of genewise deviances when estimating
        the hyperparameters. Positive values produce robust empirical Bayes
        ignoring outlier small or large deviances. Only used when
        :code:`robust=True`.

    Returns
    -------
    DGEGLM
        object with the same components as produced by :func:`glmFit`, plus:

        - :code:`df_residual_zeros`, an array containing the number of
          effective residual degrees of freedom for each gene, taking into
          account any treatment groups with all zero counts.
        - :code:`df_prior`, a float (if :code:`robust=False`) or array (if
          :code:`robust=True`), giving the prior degrees of freedom for the QL
          dispersions.
        - :code:`var_prior`, a float (if :code:`robust=False`) or array (if
          :code:`robust=True`), giving the location of the prior distrbution
          for the QL dispersions.
        - :code:`var_post`, an array containing the posterior empirical Bayes
          QL dispersions.
    """
    glmfit = glmFit(
        y,
        design=design,
        dispersion=dispersion,
        offset=offset,
        lib_size=lib_size,
        weights=weights,
    )

    # Setting up the abundances.
    if abundance_trend:
        if AveLogCPM is None:
            AveLogCPM = aveLogCPM(
                y, lib_size=lib_size, weights=weights, dispersion=dispersion
            )
        glmfit.AveLogCPM = AveLogCPM
    else:
        AveLogCPM = None

    # Adjust df_residual for fitted values at zero
    zerofit = (glmfit.fitted_values < 1e-4) & (glmfit.counts < 1e-4)
    df_residual = _residDF(zerofit, glmfit.design)

    # Empirical Bayes squeezing of the quasi-likelihood variance factors
    with np.errstate(invalid="ignore"):
        s2 = glmfit.deviance / df_residual
    s2[df_residual == 0] = 0
    s2 = np.maximum(s2, 0)
    s2_fit = squeezeVar(
        s2,
        df=df_residual,
        covariate=AveLogCPM,
        robust=robust,
        winsor_tail_p=winsor_tail_p,
    )

    # Storing results
    glmfit.df_residual_zeros = df_residual
    glmfit.df_prior = s2_fit["df_prior"]
    glmfit.var_post = s2_fit["var_post"]
    glmfit.var_prior = s2_fit["var_prior"]
    return glmfit


def glmQLFTest(glmfit, coef=None, contrast=None, poisson_bound=True):
    """
    Conduct genewise statistical tests for a given coefficient or contrast

    Implement one of the quasi-likelihood (QL) methods of [Lund2012]_, with some
    enhancements and with slightly different glm, trend and FDR methods. See
    [Lun2016]_ or [Chen2016]_ for tutorials describing the use of
    :func:`glmQLFit` and :func:`glmQLFTest` as part of a complete pipeline.
    Another case study using :func:`glmQLFit` and :func:`glmQLFTest` is given in
    Section 4.7 of the edgeR User's Guide.

    :func:`glmQLFTest` is similar to :func:`glmLRT` except that it replaces
    likelihood ratio tests with empirical Bayes quasi-likelihood F-tests. The
    *p*-values from :func:`glmQLFTest` are always greater than or equal to those
    that would be obtained from :func:`glmLRT` using the same negative binomial
    dispersions.

    Note
    ----
    The negative binomial dispersions :code:`dispersion` supplied to
    :func:`glmQLFit` and :func:`glmQLFTest` must be based on a global model, that
    is, they must be either trended or common dispersions. It is not correct to
    supply genewise dispersions because :func:`glmQLFTest` estimates genewise
    variability using the QL dispersion.

    Arguments
    ---------
    glmfit : DGEGLM
        a :class:`DGEGLM` object, usually output from :func:`glmQLFit`
    coeff : int or string array
        indicated which coefficients of the linear model are to be tested equal to zero.
        Ignored if :code:`contrast` is not :code:`None`.
    contrast : array_like
        vector or matrix specifying one or more contrasts of the linear model
        coefficients to be tested equal to zero.
    poisson_bound : bool
        if :code:`True` then the *p*-value returned will never be less than
        would be obtained for a likelihood ratio test with NB dispersion equal
        to zero.

    Returns
    -------
    DGELRT
        an object of class :class:`DGELRT` with the same components as produced
        by :func:`glmLRT`, except that the :code:`"LR"` column of the
        :code:`table` becomes :code:`"F"` and contains quasi-likelihood
        F-statistics. It also stored :code:`df_total`, an array containing the
        denominator degrees of freedom for the F-test, equal to :code:`df_prior
        + df_residual_zeros`.
    """

    if coef is None:
        coef = glmfit.design.shape[1] - 1

    if not isinstance(glmfit, DGEGLM):
        raise ValueError("glmfit must be a DGEGLM object produced by glmQLFit")
    if glmfit.var_post is None:
        raise ValueError("need to run glmQLFit before glmQLFTest")
    out = glmLRT(glmfit, coef=coef, contrast=contrast)

    # compute the QL F-statistic
    F_stat = out.table["LR"] / out.df_test / glmfit.var_post
    df_total = glmfit.df_prior + glmfit.df_residual_zeros
    max_df_residual = glmfit.counts.shape[1] - glmfit.design.shape[1]
    df_total = np.minimum(df_total, glmfit.counts.shape[0] * max_df_residual)

    # compute p-values from the QL F-statistic
    F_pvalue = scipy.stats.f.sf(F_stat, dfn=out.df_test, dfd=df_total)

    # Ensure is not more significant than chisquare test with Poisson variance
    if poisson_bound:
        i = _isBelowPoissonBound(glmfit)
        if i.any():
            # pois_fit = glmfit[i, :]
            pois_fit = glmFit(
                glmfit.counts[i, :],
                design=glmfit.design,
                offset=np.broadcast_to(glmfit.offset, glmfit.counts.shape)[i, :],
                weights=glmfit.weights[i, :] if glmfit.weights is not None else None,
                start=glmfit.unshrunk_coefficients[i],
                dispersion=0,
            )
            pois_res = glmLRT(pois_fit, coef=coef, contrast=contrast)
            F_pvalue[i] = np.maximum(F_pvalue[i], pois_res.table["PValue"])

    del out.table["LR"]
    out.table["F"] = F_stat
    out.table["PValue"] = F_pvalue
    out.df_total = df_total

    return out


def _isBelowPoissonBound(glmfit):
    """a convenience function"""
    disp = glmfit.dispersion
    s2 = glmfit.var_post[:, None]
    fitted = glmfit.fitted_values

    return (((fitted * disp + 1) * s2) < 1).any(axis=1)


def plotQLDisp(
    glmfit,
    xlab="Average Log2 CPM",
    ylab="Quarter-Root Mean Deviance",
    col_shrunk="red",
    col_trend="blue",
    col_raw="black",
):
    """
    Plot the genewise quasi-likelihood dispersion against the gene abundance (in log2 counts per million)

    This function displays the quarter-root of the quasi-likelihood dispersions
    for all genes, before and after shrinkage towards a trend. If
    :code:`glmfit` was constructed without an abundance trend, the function
    instead plots a horizontal line (of color :code:`col_trend`) at the common
    value towards which dispersions are shrunk. The quarter-root transformation
    is applied to improve visibility for dispersions around unity.

    Arguments
    ---------
    glmfit : DGEGLM
        a DGEGLM object produced by :func:`glmQLFit`
    xlab : str
        label for the x-axis
    ylab : str
        label for the y-axis
    col_shrunk : str
        color of the points representing the squeezed quasi-likelihood dispersions
    col_trend : str
        color of the line showing dispersion trend
    col_raw : str
        color of the line showing the unshrunk dispersions
    """
    import matplotlib.pyplot as plt

    A = glmfit.AveLogCPM
    if A is None:
        A = glmfit.aveLogCPM()
    A = np.asarray(A)
    s2 = glmfit.deviance / glmfit.df_residual_zeros
    if glmfit.var_post is None:
        raise ValueError("need to run glmQLFit before plotQLDisp")

    plt.scatter(A, np.sqrt(np.sqrt(s2)), c=col_raw, label="Raw")
    plt.scatter(A, np.sqrt(np.sqrt(glmfit.var_post)), c=col_shrunk, label="Shrunk")
    if isinstance(glmfit.var_prior, float):
        plt.axhline(y=np.sqrt(np.sqrt(glmfit.var_prior)), c=col_trend, label="Trend")
    else:
        o = np.argsort(A)
        plt.plot(
            A[o], np.sqrt(np.sqrt(glmfit.var_prior[o])), c=col_trend, label="Trend"
        )

    plt.legend(loc="upper right")
    plt.show()



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2024 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/q2qnbinom.R' of the Bioconductor edgeR package (version 3.38.4).

import numpy as np
from .edgepy_cpp import _q2qnbinom


# the actual implementation in done in Cython
# this Python wrapper allows to use optional and named arguments
def q2qnbinom(x, input_mean, output_mean, dispersion=0):
    """
    Interpolated quantile to quantile mapping between negative-binomial distributions with the same dispersion but different means.

    This function finds the quantile with the same left and right tail
    probabilities relative to the output mean as :code:`x` has relative to the
    input mean. In principle, :func:`q2qnbinom` gives similar results to calling
    :code:`pnbinom` followed by :code:`qnbinom` as in the example below.
    However this function avoids infinite values arising from rounding errors
    and does appropriate interpolation to return continuous values.

    It is called by :func:`equalizeLibSizes` to perform quantile-to-quantile
    normalization.

    See also
    --------
    equalizeLibSizes

    Arguments
    ---------
    x : array_like
        matrix of counts
    input_mean : array_like
        matrix of population means for :code:`x`. If 1-D, then of the same
        length as :code:`x.shape[0]`
    output_mean : array_like
        matrix of population means for the output values. If 1-D, then of the
        same length as :code:`x.shape[0]`
    dispersion : array_like
        scalar, vector or matrix giving negative binomial dispersion values

    Returns
    -------
    ndarray
        matrix of same dimensions as :code:`x`, with :code:`output_mean` as the
        new nominal population mean
    """

    return _q2qnbinom(x, input_mean, output_mean, dispersion)



import unittest
import numpy as np

from inmoose.limma import fitFDist
from inmoose.limma.fitFDist import trigammaInverse


class Test(unittest.TestCase):
    def test_trigammainverse(self):
        res = trigammaInverse(np.array([np.nan, 0, 1e8, -1, 5]))
        self.assertTrue(
            np.allclose(
                res,
                [np.nan, np.inf, 0.0001000, np.nan, 0.4961687],
                equal_nan=True,
                atol=1e-5,
            )
        )



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2024 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/exactTestBetaApprox.R' of the Bioconductor edgeR package (version 3.38.4).

import numpy as np
from scipy.stats import beta


def exactTestBetaApprox(y1, y2, dispersion=0.0):
    """
    Compute genewise *p*-values for differences in the means between two groups of negative-binomially distributed counts.

    This function implements an asymptotic beta distribution approximation to
    the conditional count distribution.

    See also
    --------
    exactTest

    Arguments
    ---------
    y1 : matrix
        matrix of counts for the first of the two experimental groups to be
        tested for differences. Rows correspond to genes and columns to
        libraries. Libraries are assumed to be equal in size -- *e.g.* adjusted
        pseudocounts from the output of :func:`equalizeLibSizes`.
    y2 : matrix
        matrix of counts for the second of the two experimental groups to be
        tested for differences. Rows correspond to genes and columns to
        libraries. Libraries are assumed to be equal in size -- *e.g.* adjusted
        pseudocounts from the output of :func:`equalizeLibSizes`.
    dispersion : array_like of floats
        an array of dispersions, either of length one or of length equal to the
        number of genes.

    Returns
    -------
    ndarray
        array of genewise *p*-values, one for each row of :code:`y1` and :code:`y2`
    """
    y1 = np.asarray(y1)
    y2 = np.asarray(y2)

    # convert matrices to vectors
    ntags = y1.shape[0]
    n1 = y1.shape[1]
    n2 = y2.shape[1]
    y1 = np.round(y1.sum(axis=1))
    y2 = np.round(y2.sum(axis=1))

    dispersion = np.broadcast_to(dispersion, (ntags,))

    # Null fitted values
    y = y1 + y2
    mu = y / (n1 + n2)

    # Compute p-values
    pvals = np.ones(ntags)
    all_zero = y <= 0
    alpha1 = n1 * mu / (1 + dispersion * mu)
    alpha2 = n2 / n1 * alpha1
    med = np.zeros(ntags)
    med[~all_zero] = beta.ppf(0.5, alpha1[~all_zero], alpha2[~all_zero])
    left = ((y1 + 0.5) / y < med) & ~all_zero
    if left.any():
        pvals[left] = 2 * beta.cdf(
            (y1[left] + 0.5) / y[left], alpha1[left], alpha2[left]
        )
    right = ((y1 - 0.5) / y > med) & ~all_zero
    if right.any():
        pvals[right] = 2 * beta.sf(
            (y1[right] - 0.5) / y[right], alpha1[right], alpha2[right]
        )
    return pvals



# -----------------------------------------------------------------------------
# Copyright (C) 2004-2022 Gordon Smyth, Yifang Hu, Matthew Ritchie, Jeremy Silver, James Wettenhall, Davis McCarthy, Di Wu, Wei Shi, Belinda Phipson, Aaron Lun, Natalie Thorne, Alicia Oshlack, Carolyn de Graaf, Yunshun Chen, Mette Langaas, Egil Ferkingstad, Marcus Davy, Francois Pepin, Dongseok Choi
# Copyright (C) 2024 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/squeezeVar.R' of the Bioconductor limma package (version 3.55.1).


import numpy as np

from .fitFDist import fitFDist


def squeezeVar(var, df, covariate=None, robust=False, winsor_tail_p=(0.05, 0.1)):
    """
    Squeeze a set of sample variances together by computing empirical Bayes posterior means
    This function implements empirical Bayes algorithms proposed by
    [Smyth2004]_ and [Phipson2016]_.

    A conjugate Bayesian hierarchical model is assumed for a set of sample
    variances. The hyperparameters are estimated by fitting a scaled
    F-distribution to the sample variances. The function returns the posterior
    variances and the estimated hyperparameters.

    Specifically, the sample variances :code:`var` are assumed to follow scaled
    chi-squared distributions, conditional on the true variances, and a scaled
    inverse chi-squared prior is assumed for the true variances.  The scale and
    degrees of freedom of this prior distribution are estimated from the values
    of :code:`var`.

    The effect of this function is to squeeze the variances towards a common
    value, or to a global trend if a :code:`covariate` is provided. The
    squeezed variances have a smaller expected mean square error to the true
    variances than do the sample variances themselves.

    If :code:`covariate` is not :code:`None`, then the scale parameter of the
    prior distribution is assumed to depend on the covariate. If the covariate
    is average log-expression, then the effect is an intensity-dependent trend
    similar to that in [Sartor2006]_.

    :code:`robust=True` implements the robust empirical Bayes procedure of
    [Phipson2016]_ which allows some of the :code:`var` values to be outliers.

    Arguments
    ---------
    var : array_like
        1-D array of independent sample variances
    df : array_like
        1-D array of degrees of freedom for the sample variances
    covariate : ??
        if not :code:`None`, :code:`var_prior` will depend on this numeric
        covariate. Otherwise, :code:`var_prior` is constant.
    robust : bool
        whether the estimation of :code:`df_prior` and :code:`var_prior` be
        robustified against outlier sample variances
    winsor_tail_p : float or pair of floats
        left and right tail proportions of :code:`x` to Winsorize. Only used
        when :code:`robust=True`

    Returns
    -------
    dict
        a dictionnary with keys:

        - :code:`"var_post"`, 1-D array of posterior variances. Same length as
          :code:`var`.
        - :code:`"var_prior"`, location or scale of prior distribution. 1-D
          array of same length as :code:`var` if :code:`covariate` is not
          :code:`None`, otherwise a single value.
        - :code:`"df_prior"`, degrees of freedom of prior distribution. 1-D
          array of same length as :code:`var` if :code:`robust=True`, otherwise
          a single value.
    """
    n = len(var)

    # Degenerate special cases
    if n == 0:
        raise ValueError("var is empty")
    if n == 1:
        return {"var_post": var, "var_prior": var, "df_prior": 0}

    # When df==0, guard against missing or infinite values in var
    df = np.broadcast_to(df, var.shape)
    var[df == 0] = 0

    # Estimate hyperparameters
    if robust:
        raise NotImplementedError("Robust estimation in squeezeVar is not implemented")
    else:
        fit = fitFDist(var, df1=df, covariate=covariate)
        df_prior = fit["df2"]
    if np.isnan(df_prior).any():
        raise RuntimeError("Could not estimate prior df")

    # Posterior variances
    var_post = _squeezeVar(var=var, df=df, var_prior=fit["scale"], df_prior=df_prior)

    return {"var_post": var_post, "var_prior": fit["scale"], "df_prior": df_prior}


def _squeezeVar(var, df, var_prior, df_prior):
    """
    Squeeze posterior variances given hyperparameters

    NAs not allowed in :code:`df_prior`

    Arguments
    ---------
    var : array_like
        1-D array of independent sample variances
    df : array_like
        1-D array of degrees of freedom for the sample variances
    var_prior : array_like
        array of prior variances
    df_prior :
        array of degrees of freedom for the prior variances

    Returns
    -------
    ndarray
        array of posterior variances
    """
    df = np.broadcast_to(df, var.shape)
    var_prior = np.broadcast_to(var_prior, var.shape)
    df_prior = np.broadcast_to(df_prior, var.shape)

    isfin = np.isfinite(df_prior)
    var_post = np.zeros(var.shape)

    # For infinite df_prior, return var_prior
    var_post[~isfin] = var_prior[~isfin]

    var_post[isfin] = (df[isfin] * var[isfin] + df_prior[isfin] * var_prior[isfin]) / (
        df[isfin] + df_prior[isfin]
    )

    return var_post



import unittest
import numpy as np

from inmoose.utils import rnbinom
from inmoose.edgepy import nbinomDeviance


class test_nbinomDeviance(unittest.TestCase):
    def setUp(self):
        y = np.array(rnbinom(80, size=5, mu=20, seed=42)).reshape((20, 4))
        self.y = np.vstack(([0, 0, 0, 0], [0, 0, 2, 2], y))

    def test_nbinomDeviance(self):
        dev_ref = np.array(
            [
                7.999999,
                5.545177,
                321.818901,
                184.761910,
                355.515748,
                340.390858,
                414.784086,
                461.324010,
                235.746631,
                354.718673,
                410.643778,
                265.824737,
                422.344282,
                412.896876,
                392.316278,
                599.686176,
                242.088733,
                501.330767,
                608.676351,
                706.581453,
                430.474599,
                310.504051,
            ]
        )
        dev = nbinomDeviance(self.y, np.ones(self.y.shape))
        self.assertTrue(np.allclose(dev, dev_ref, atol=0, rtol=10 - 9))

        dev_ref = np.array([8.454617])
        dev = nbinomDeviance(np.array([1, 2, 3, 4]), np.ones(4), dispersion=np.zeros(4))
        self.assertTrue(np.allclose(dev, dev_ref, atol=0, rtol=10 - 9))

        with self.assertRaisesRegex(
            ValueError, expected_regex="y is a matrix but mean is not"
        ):
            nbinomDeviance(self.y, np.ones((22,)))

        with self.assertRaisesRegex(
            ValueError, expected_regex="mean is a matrix but y is not"
        ):
            nbinomDeviance(np.array([1, 2, 3, 4]), np.ones((4, 3)))

        with self.assertRaisesRegex(
            ValueError, expected_regex="length of mean differs from that of y"
        ):
            nbinomDeviance(np.array([1, 2, 3, 4]), np.ones(3))
